#define MAX 80 
#define PORT 8000 
#define SA struct sockaddr
#include <stdio.h> 
#include <pthread.h>
#include <unistd.h>
#include <sys/types.h> 
#include <arpa/inet.h>
#include <arpa/inet.h>
#include <netinet/in.h> 
#include <stdlib.h>
#include "KVClientLibrary.h"
#include <string.h>

struct message* requestMessage;

struct config {
    int listening_port;
    int cache_size;
    int thread_pool_size;
};

struct config *settings;

int read_config(struct config *settings) {
    FILE * fp;
    char * line = NULL;
    size_t len = 0;
    ssize_t read;
    char *param;
    char *value;
    int i;

    fp = fopen("settings.conf", "r");
    if (fp == NULL)
        return -1;

    while ((read = getline(&line, &len, fp)) != -1) {
        i = 0;
        if(line[0]=='#') {
            break;
        } else {
            param=strtok(line,"=");
            value=strtok(NULL,"=");
            if(strcmp(param,"listening_port")==0) {
                settings->listening_port = atoi(value);
            } else if (strcmp(param,"cache_size")==0) {
                settings->cache_size = atoi(value);
            } else if (strcmp(param,"thread_pool_size")==0) {
                settings->thread_pool_size = atoi(value);
            } else {

            }
        }
    }

    fclose(fp);
}


void interactive (int sock) {
    char command[515];
    int readlen, i, j;
    int status_code;
    
    char key[256];
    char value[256];
    char error[256];

    do {
        memset(key, 0, sizeof(key)); 
        memset(value, 0, sizeof(value)); 
        memset(error, 0, sizeof(error)); 
        printf("interactive> ");
        fflush(stdout);
        readlen = read(0, command, sizeof(command));
        i=5;
        while(command[i] && command[i]!=' ' && command[i]!='\n') {
            i++;
        }
        memcpy(key, command+4, i-4);
        if(command[i]==' ') {
            j=i+1;
            while(command[j] && command[j]!=':' && command[j]!='\n') {
                j++;
            }
            memcpy(value, command+i+1, j-i-1);
        }
        
        // printf("key: %s\n", key);
        // printf("value: %s\n", value);
        // printf("error: %s\n", error);

        switch(command[0]) {
            case 'g': status_code = get(sock, key, value, error);
                printf("[%d]", status_code);
                if(status_code==200) {
                    printf(" Value recieved: %s\n", value);
                } else {
                    printf(" Error: %s\n", error);
                }
                break;
            case 'p':status_code = put(sock, key, value, error);
                printf("[%d]\n", status_code);
                break;
            case 'd':status_code = del(sock, key, error);
                printf("[%d]", status_code);
                if(status_code==200) {
                    printf(" Deleted successfully %s\n", value);
                } else {
                    printf(" Error: %s\n", error);
                }
                break;
            default: return;
        }

    } while(command[0] != 'q');
}

void gen_random(char *s, const int len) {
    static const char alphanum[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";

    for (int i = 0; i < len; ++i) {
        s[i] = alphanum[rand() % (sizeof(alphanum) - 1)];
    }

    s[len] = 0;
}


int client_init() {
    int valread;
    struct sockaddr_in serv_addr; 
    char *hello = "Hello from client"; 
    char buffer[1024] = {0};  
    int sock = 0;
    if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) 
    { 
        printf("\n Socket creation error \n"); 
        return -1; 
    } 
   
    serv_addr.sin_family = AF_INET; 
    serv_addr.sin_port = htons(settings->listening_port); 
       
    // Convert IPv4 and IPv6 addresses from text to binary form 
    if(inet_pton(AF_INET, "127.0.0.1", &serv_addr.sin_addr)<=0)  
    { 
        printf("\nInvalid address/ Address not supported \n"); 
        return -1; 
    } 
   
    if (connect(sock, (struct sockaddr *)&serv_addr, sizeof(serv_addr)) < 0) 
    { 
        printf("\nConnection Failed \n"); 
        return -1; 
    } 

    return sock;
}

void *seige(void * args) {

    int *rcompleted = (int *) args;
    *rcompleted = 0;
    int status_code;
    char key[256];
    char value[256];
    char error[256];
    int sock;
    int r;

    sock = client_init();

    while(1) {
        gen_random(key, 10);
        gen_random(value, 10);
        r = rand() % 3;
        switch(r) {
            case 0: put(sock, key, value, error);
                break;
            case 1: get(sock, key, value, error);
                break;
            case 2: get(sock, key, value, error);
                break;
        }
        *rcompleted = *rcompleted + 1;
    }

}
int main(int argc, char const *argv[]) 
{ 
    int sock;
    int max_concurrent = 100;
    int n = 2;
    pthread_t *seige_threads;
    int *requestcompleted;
    int *requestsent;
    int sum;
    int seconds = 1;
    int i;
    int seige_mode = 0;

    if(argc == 2) {
        seige_mode = 1;
    }

    settings = (struct config *) malloc(sizeof(struct config));
    if(read_config(settings)<0) {
        printf("Invalid config file!\n");
        return -1;
    }

    if(seige_mode==0) {
        sock = client_init();
        printf("To send GET message type: get key\n");
        printf("To send PUT message type: put key value\n");
        printf("To send DEL message type: del key\n");
        interactive(sock);
        return 0; 
    } else {
        seige_threads = malloc(max_concurrent * sizeof(pthread_t));
        requestcompleted = (int *) malloc(max_concurrent * sizeof(int));

        for(i=0; i<n; i++) {
            pthread_create( &seige_threads[i], NULL, seige, requestcompleted+i );
        }
        while(1) {
            sleep(1);
            sum = 0;
            for(int i=0; i<n; i++) {
                sum = sum + requestcompleted[i];
            }
            if(seconds%5==0) {
                printf("%d, %f\n", i, ((float)sum/seconds));
                pthread_create( &seige_threads[i], NULL, seige, requestcompleted+i );
                i++;
            }
            if(i==max_concurrent) {
                break;
            }
            seconds++;
        }
    }
} 