#include <stdio.h>
#include <pthread.h>
#include <stdlib.h>
#define _GNU_SOURCE
#include <fcntl.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/epoll.h>
#include <string.h> 
#include "LRU.h"
#include "StorageHandler.h"
#include "KVMessageFormat.h"
#define MAX_EVENTS 10
#define DEBUG (0)

struct config {
    int listening_port;
    int cache_size;
    int thread_pool_size;
};

void *worker(void *args) {
    struct epoll_event ev,events[MAX_EVENTS];
    struct epoll_event socketEvent;
    socketEvent.events=EPOLLIN;

    int read_pipe, conn_sock, nfds, i;
    int epollfd;
    int newfd;
    int status;
    
    epollfd = epoll_create1(0);
    while (epollfd == -1) {
        epollfd = epoll_create1(0);
    }

    /* Add the pipe which we get from main thread */
    read_pipe = ((int *) args)[0];   
    ev.events = EPOLLIN;
    ev.data.fd = read_pipe;
    if (epoll_ctl(epollfd, EPOLL_CTL_ADD, read_pipe, &ev) == -1) {
        perror("epoll_ctl: read_pipe");
        exit(EXIT_FAILURE);
    }

    while (1) {

        nfds = epoll_wait(epollfd, events, MAX_EVENTS, -1);
        if (nfds == -1) {
            perror("epoll_wait");
            exit(EXIT_FAILURE);
        }

        for ( i= 0; i < nfds; ++i ) {
            if (events[i].data.fd == read_pipe) {
                /*  if we get a request from main thread to add new client */
                read(read_pipe, &newfd, sizeof(newfd));
                printf("[Worker thread] New client assigned at socket %d\n", newfd);
                ev.data.fd=newfd;
                ev.events = EPOLLIN | EPOLLRDHUP;
                if (epoll_ctl(epollfd, EPOLL_CTL_ADD, newfd,  &ev) == -1) {
                    perror("epoll_ctl: read_pipe");
                    exit(EXIT_FAILURE);
                }

               
            } else {
                /*  if we get a request from client (GET, PUSH, DEL) */
                int flag = events[i].events;

                if (flag & EPOLLRDHUP) {
                    /* Connection was closed. */
                    epoll_ctl( epollfd, EPOLL_CTL_DEL, events[i].data.fd , NULL );
                    close(events[i].data.fd);
                    continue; 
                }

                if (flag & EPOLLIN) {
                    /* Parse the actual message from client */
                    struct message *requestMessage= malloc(sizeof(struct message));    
                    memset(requestMessage->key, 0, 256);           
                    memset(requestMessage->value, 0, 256);
                    int readlength=read(events[i].data.fd  , requestMessage, sizeof(struct message));    
                    switch(requestMessage->status) {
                        case STATUS_GET:
                            printf("[GET][Socket:%d] Key:%s\n", newfd, requestMessage->key);
                            status = cache_get(requestMessage->key, requestMessage->value);
                            if(status) {
                                requestMessage->status = 200;
                            } else {
                                status = file_get(requestMessage->key, requestMessage->value);
                                if(status) {
                                    cache_put(requestMessage->key, requestMessage->value);
                                    requestMessage->status = 200;
                                } else {
                                    requestMessage->status = 240;
                                }
                            }
                            break;
                        case STATUS_PUT:
                            printf("[PUT][Socket:%d] Key:%s Value:%s\n", newfd, requestMessage->key, requestMessage->value);
                            cache_put(requestMessage->key, requestMessage->value);
                            file_put(requestMessage->key, requestMessage->value);
                            requestMessage->status = 200;
                            break;
                        case STATUS_DEL:
                            printf("[DEL][Socket:%d] Key:%s\n", newfd, requestMessage->key);
                            status = cache_del(requestMessage->key);
                            status = file_del(requestMessage->key);
                            if(status) {
                                requestMessage->status = 200;
                            } else {
                                requestMessage->status = 240;
                            }
                            break;
                    }
                    

                    write(events[i].data.fd, requestMessage, sizeof(struct message));
                    free(requestMessage);
                }                                 
               
            }
        }
        
    }
    
}

int read_config(struct config *settings) {
    FILE * fp;
    char * line = NULL;
    size_t len = 0;
    ssize_t read;
    char *param;
    char *value;
    int i;

    fp = fopen("settings.conf", "r");
    if (fp == NULL)
        return -1;

    while ((read = getline(&line, &len, fp)) != -1) {
        i = 0;
        if(line[0]=='#') {
            break;
        } else {
            param=strtok(line,"=");
            value=strtok(NULL,"=");
            if(strcmp(param,"listening_port")==0) {
                settings->listening_port = atoi(value);
            } else if (strcmp(param,"cache_size")==0) {
                settings->cache_size = atoi(value);
            } else if (strcmp(param,"thread_pool_size")==0) {
                settings->thread_pool_size = atoi(value);
            } else {

            }
        }
    }

    fclose(fp);
}

int main (int argc, int argv) {

    int i;
    int next_thread_to_assign = 0;
    pthread_t *worker_threads;
    int sockfd, newsockfd, portno, clilen, n;
    struct sockaddr_in serv_addr, cli_addr; 

    struct config *settings;
    settings = (struct config *) malloc(sizeof(struct config));
    if(read_config(settings)<0) {
        printf("Invalid config file!\n");
        return -1;
    }

    //int *pipes = (int*) malloc(pool_thread_size * 2 * sizeof(pipes));
    int pipes[50][2];                                // TODO: initialize pipes dynamically

    worker_threads = malloc(settings->thread_pool_size * sizeof(pthread_t));
    for( i=0; i < settings->thread_pool_size; i++ ) {
        pipe(pipes[i]);
        pthread_create( &worker_threads[i], NULL, worker, &pipes[i]);
    }

    sockfd = socket(AF_INET, SOCK_STREAM, 0); 
    if (sockfd < 0) {
        perror("ERROR opening socket");
        exit(1);
    }
    memset(&serv_addr, 0, sizeof(serv_addr));
    portno = settings->listening_port;
    serv_addr.sin_family = AF_INET;
    serv_addr.sin_port = htons(portno);
    serv_addr.sin_addr.s_addr = INADDR_ANY;
    int option =1;
    setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &option, sizeof(option));
    if (bind(sockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0) {
        perror("ERROR on binding");
        exit(1);
    }
    
    listen(sockfd, 5);
    clilen = sizeof(cli_addr);

    printf("Listening on port %d\n", settings->listening_port);

    init_cache(settings->cache_size);
    storage_init();

    while(1) {

        newsockfd = accept(sockfd, (struct sockaddr *) &cli_addr, &clilen);
        if (newsockfd < 0) 
            perror("ERROR on accept");

        write(pipes[next_thread_to_assign][1], &newsockfd, sizeof(newsockfd));
        next_thread_to_assign = (next_thread_to_assign+1) % settings->thread_pool_size;
    } 

    return 0;
}
