#include <stdio.h>
#include <pthread.h>
#include <stdlib.h>
#define _GNU_SOURCE
#include <fcntl.h>
#include <unistd.h>
#include <string.h> 
#include <stdatomic.h> 
#include <assert.h>
#include "LRU.h"
#define ATOMIC_TEST_AND_SET  __atomic_test_and_set
#define CLEAR __atomic_clear


struct KV **array;
struct queue *qu=NULL;
struct queue *last;
unsigned _Atomic lock=0;
int maxsize;


void remove_element_from_deque(char *key)
{
    while(ATOMIC_TEST_AND_SET(&(lock),1) == 1);
    queue *present = qu , *previous=NULL;
    if(present == NULL){
        CLEAR(&(lock),0);
        return;
    }
    
    if(strcmp(present->key, key) == 0)
    {
        qu = qu->next;
        if(last == present)
            last = last->next;
        free(present);
        CLEAR(&(lock),0);
        return;
    }

    while(present->next != NULL)
    {
        if(strcmp(present->key, key) == 0)
        {
            if(last == present)
                last = previous;
            previous->next = present->next;
            free(present);
            CLEAR(&(lock),0);
            return;
        }
        previous = present;
        present = present->next;
    }
    CLEAR(&(lock),0);
}

void insert_into_queue(char *key)
{
    while(ATOMIC_TEST_AND_SET(&(lock),1) == 1);
    queue *temp = (queue *)malloc(sizeof(queue));
    memcpy(temp->key, key, KEY_SIZE);
    temp->next = NULL;
    if(qu == NULL)
    {
        qu = temp;
        last = temp;
    }
    else
    {
        last->next = temp;
        last = temp;
    }
    CLEAR(&(lock),0);
}

int remove_front_element()
{
    while(ATOMIC_TEST_AND_SET(&(lock),1) == 1);
    queue *temp = qu;
    qu = qu->next;
    assert(temp != NULL);
    assert(qu != NULL);
    for(int i=0;i<maxsize;i++)
    {
        if(strcmp(array[i]->key , temp->key) == 0)
        {
            // check if it is modified to write to file
            array[i]->valid = FALSE;
            CLEAR(&(lock),0);
            return i;
        }
    }
    CLEAR(&(lock),0);
}

int find_empty_location()
{
    for(int i=0;i<maxsize;i++)
    {
        if(array[i]->valid == FALSE)
        {
            CLEAR(&(lock),0);
            return i;
        }
    }
    int indx = remove_front_element();
    return indx;
}

int cache_del(char *key)
{
    for(int i=0;i<maxsize;i++)
    {
        if(strcmp(array[i]->key , key) == 0)
        {
            while(ATOMIC_TEST_AND_SET(&(array[i]->lock),1) == 1);
            remove_element_from_deque(key);
            array[i]->valid = FALSE;
            CLEAR(&(array[i]->lock),0);
            return 1;
        }
    }
    return 0;
    //TODO remove key from file also
}

void cache_put(char *key, char *value)
{
    int indx=-1;
    for(int i=0;i<maxsize;i++)
    {
        if(array[i]->key!=NULL) {
            if(strcmp(array[i]->key , key) == 0)
            {
                indx = i;
                remove_element_from_deque(key);
                break;
            }
        }
    }

    if(indx == -1)
    {
        indx = find_empty_location();
        assert(indx>=0);
        // TODO should write to file if modified is true
        // replacment from cache
    }

    while(ATOMIC_TEST_AND_SET(&(array[indx]->lock),1) == 1);
    memcpy(array[indx]->key, key, KEY_SIZE);
    memcpy(array[indx]->value, value, VAL_SIZE);
    array[indx]->valid = TRUE;
    array[indx]->modified = TRUE;
    insert_into_queue(key);
    CLEAR(&(array[indx]->lock),0);
}

int cache_get(char *key, char *value)
{
    for(int i=0;i<maxsize;i++)
    {
        if(strcmp(array[i]->key , key) == 0 && array[i]->valid)
        {
            while(ATOMIC_TEST_AND_SET(&(array[i]->lock),1) == 1);
            remove_element_from_deque(key);
            insert_into_queue(key);
            CLEAR(&(array[i]->lock),0);
            memcpy(value, array[i]->value, VAL_SIZE);
            return 1;
        }
    }
    return 0;
}

void init_cache(int size) {
    maxsize = size;
    array = (struct KV **) malloc(maxsize * sizeof(struct KV));
    for(int j=0;j<maxsize;j++)
    {
        array[j] = (KV *)malloc(sizeof(KV));
        array[j]->valid = FALSE;
        array[j]->lock = 0;
    }
}

void print_cache() {
    for(int j=0;j<maxsize;j++)
    {
        printf("[%d] (%s) : (%s)\n", array[j]->valid, array[j]->key, array[j]->value);
    }
}
